<?php
/* --------------------------------------------------------------
	CleverReachSync.inc.php 2016-04-11
	Gambio GmbH
	http://www.gambio.de
	Copyright (c) 2015 Gambio GmbH
	Released under the GNU General Public License (Version 2)
	[http://www.gnu.org/licenses/gpl-2.0.html]
	--------------------------------------------------------------
*/

class CleverReachSync
{
	protected $configStorage;
	protected $db;
	protected $logger;

	public function __construct()
	{
		$this->configStorage = MainFactory::create('CleverReachConfigurationStorage');
		$this->db = StaticGXCoreLoader::getDatabaseQueryBuilder();
		$this->logger = MainFactory::create('CleverReachLogger');
	}

	public function getNumberOfQueueEntries()
	{
		$numEntries = $this->db->select('id')->where('status', '0')->from('cleverreach_api_sync')->get()->num_rows();
		return $numEntries;
	}

	/**
	 * syncs receivers upstream (i.e. shop -> cr)
	 * @param  int $id primary key from cleverreach_api_sync table (sync single entry)
	 * @return array     ids of entries processed
	 */
	public function syncUpstreamQueue($id = null, CleverReachGroup $group = null)
	{
		if ($group === null)
		{
			$group = MainFactory::create('CleverReachGroup', $this->configStorage->get('newsletter_group_id'));
		}
		$maxBulkReceivers = 50;
		$customerReadService = StaticGXCoreLoader::getService('CustomerRead');
		$this->db->where('status', '0');
		if($id !== null)
		{
			$this->db->where('id', (int)$id);
		}
		$query = $this->db->get('cleverreach_api_sync');
		$receivers = array();
		$syncedReceiversIds = array();
		foreach($query->result() as $queueRow)
		{
			if($queueRow->action === 'ADD')
			{
				try
				{
					$customer = $customerReadService->getCustomerById(MainFactory::create('IdType', $queueRow->customers_id));
					$defaultAddress = $customer->getDefaultAddress();
					$newReceiverData = array(
						'email' => $queueRow->email,
						'global_attributes' => array(
							'firstname'  => (string)$customer->getFirstname(),
							'lastname'   => (string)$customer->getLastname(),
							'street'     => (string)$defaultAddress->getStreet(),
							'zip'        => (string)$defaultAddress->getPostcode(),
							'city'       => (string)$defaultAddress->getCity(),
							'country'    => (string)$defaultAddress->getCountry()->getName(),
							'company'    => (string)$defaultAddress->getCompany(),
							'gender'     => (string)$customer->getGender(),
							'dob'        => $customer->getDateOfBirth()->format('Y-m-d'),
							'date_added' => $this->getCustomerDateAdded($customer)->format('Y-m-d'),
						),
					);
				}
				catch (InvalidArgumentException $e)
				{
					$this->logger->notice(sprintf('Entry %d in queue references invalid customer with ID %d, transmitting email address only',
					  $queueRow->id, $queueRow->customers_id));
					$newReceiverData = array(
						'email' => $queueRow->email,
					);
				}
				$receivers[] = $newReceiverData;
				$syncedReceiversIds[] = $queueRow->id;
				if(count($receivers) >= $maxBulkReceivers)
				{
					$group->addReceivers($receivers);
					$receivers = array();
					$this->db->set('status', 1)
					         ->set('updated_at', date('Y-m-d H:i:s'))
					         ->where('id IN (' . implode(',', $syncedReceiversIds) . ')')
					         ->update('cleverreach_api_sync');
					$syncedReceiversIds = array();
				}
			}
			if($queueRow->action === 'DEL')
			{
				$group->deleteReceiver($queueRow->email);
				$this->db->set('status', 1)
				         ->set('updated_at', date('Y-m-d H:i:s'))
				         ->where('id', $queueRow->id)
				         ->update('cleverreach_api_sync');
			}
		}
		if(!empty($receivers))
		{
			$group->addReceivers($receivers);
			$receivers = array();
			$this->db->set('status', 1)
			         ->set('updated_at', date('Y-m-d H:i:s'))
			         ->where('id IN (' . implode(',', $syncedReceiversIds) . ')')
			         ->update('cleverreach_api_sync');
			$syncedReceiversIds = array();
		}
	}

	public function sendAllCustomers(CleverReachGroup $group)
	{
		$this->logger->notice('sending all customers to group ' . $group->name);
		$remoteReceivers = $this->getRemoteSubscribers($group);
		$remoteReceiversToDelete = $remoteReceivers;
		$chunkSize = 50;
		$limit = 50;
		$offset = 0;
		$receiversSent = 0;
		$receivers = array();
		$customerReadService = StaticGXCoreLoader::getService('CustomerRead');
		$customers = $customerReadService->filterCustomers(array(), $limit, $offset);
		while(!empty($customers))
		{
			foreach($customers as $customer)
			{
				$remoteReceiversToDelete = array_diff($remoteReceiversToDelete, array((string)$customer->getEmail()));
				if(!in_array((string)$customer->getEmail(), $remoteReceivers))
				{
					$defaultAddress = $customer->getDefaultAddress();
					$newReceiverData = array(
						'email' => (string)$customer->getEmail(),
						'global_attributes' => array(
							'firstname'  => (string)$customer->getFirstname(),
							'lastname'   => (string)$customer->getLastname(),
							'street'     => (string)$defaultAddress->getStreet(),
							'zip'        => (string)$defaultAddress->getPostcode(),
							'city'       => (string)$defaultAddress->getCity(),
							'country'    => (string)$defaultAddress->getCountry()->getName(),
							'company'    => (string)$defaultAddress->getCompany(),
							'gender'     => (string)$customer->getGender(),
							'dob'        => $customer->getDateOfBirth()->format('Y-m-d'),
							'date_added' => $this->getCustomerDateAdded($customer)->format('Y-m-d'),
						),
					);
					$receivers[] = $newReceiverData;
				}
				if(count($receivers) >= $chunkSize)
				{
					$group->addReceivers($receivers);
					$receiversSent += count($receivers);
					$receivers = array();
				}
			}
			$offset += $limit;
			$customers = $customerReadService->filterCustomers(array(), $limit, $offset);
		}

		if(!empty($receivers))
		{
			$group->addReceivers($receivers);
			$receiversSent += count($receivers);
		}

		foreach($remoteReceiversToDelete as $deleteReceiver)
		{
			$group->deleteReceiver($deleteReceiver);
		}

		return $receiversSent;
	}

	public function sendAllBuyers(CleverReachGroup $group)
	{
		$chunkSize = 50;
		$buyers = array();
		$buyersSent = array();
		$numBuyersSent = 0;
		$remoteBuyers = $this->getRemoteSubscribers($group);
		$db = StaticGXCoreLoader::getDatabaseQueryBuilder();
		$orderCustomersQuery =
			$db->select('customers_id, customers_email_address, customers_firstname, customers_lastname, customers_gender, customers_company, customers_street_address, customers_city, customers_postcode, customers_country')
			   ->get('orders');
		$customerReadService = StaticGXCoreLoader::getService('CustomerRead');
		foreach($orderCustomersQuery->result() as $row)
		{
			if(!in_array($row->customers_email_address, $remoteBuyers) && !in_array($row->customers_email_address, $buyersSent))
			{
				try
				{
					$customer   = $customerReadService->getCustomerById(MainFactory::create('IdType', $row->customers_id));
					$dob        = $customer->getDateOfBirth()->format('Y-m-d');
					$date_added = $this->getCustomerDateAdded($customer)->format('Y-m-d');
				}
				catch(InvalidArgumentException $e)
				{
					$dob = '1000-01-01';
					$date_added = '1000-01-01';
				}
				$newReceiverData = array(
					'email' => (string)$row->customers_email_address,
					'global_attributes' => array(
						'firstname'  => (string)$row->customers_firstname,
						'lastname'   => (string)$row->customers_lastname,
						'street'     => (string)$row->customers_street_address,
						'zip'        => (string)$row->customers_postcode,
						'city'       => (string)$row->customers_city,
						'country'    => (string)$row->customers_country,
						'company'    => (string)$row->customers_company,
						'gender'     => (string)$row->customers_gender,
						'dob'        => $dob,
						'date_added' => $date_added,
					),
				);
				$buyers[] = $newReceiverData;
				$buyersSent[] = (string)$row->customers_email_address;
			}
			if(count($buyers) >= $chunkSize)
			{
				$group->addReceivers($buyers);
				$numBuyersSent += count($buyers);
				$buyers = array();
			}
		}
		if(!empty($buyers))
		{
			$group->addReceivers($buyers);
			$numBuyersSent += count($buyers);
		}
		return $numBuyersSent;
	}

	public function completeNewsletterSync(CleverReachGroup $group)
	{
		$report = '';
		$localSubscribers = $this->getLocalSubscribers();
		$remoteSubscribers = $this->getRemoteSubscribers($group);

		$report .= sprintf("local subscribers:\n%s\n", print_r($localSubscribers, true));
		$report .= sprintf("\nremote subscribers:\n%s\n", print_r($remoteSubscribers, true));
		$report .= sprintf("prio: %s\n", $this->configStorage->get('prio'));
		$this->logger->notice(
			sprintf(
				'performing newsletter sync for %d local and %d remote entries, priority on %s',
				count($localSubscribers),
				count($remoteSubscribers),
				$this->configStorage->get('prio')
			)
		);

		if($this->configStorage->get('prio') === 'webshop')
		{
			$subscribersToAdd = array_diff($localSubscribers, $remoteSubscribers);
			$subscribersToDelete = array_diff($remoteSubscribers, $localSubscribers);
			$this->enqueueReceivers($subscribersToAdd, 'ADD');
			$this->enqueueReceivers($subscribersToDelete, 'DEL');
			$this->syncUpstreamQueue();
		}

		if($this->configStorage->get('prio') === 'cleverreach')
		{
			$subscribersToAdd = array_diff($remoteSubscribers, $localSubscribers);
			$subscribersToDelete = array_diff($localSubscribers, $remoteSubscribers);
			$this->activateLocalSubscribers($subscribersToAdd);
			$this->deactivateLocalSubscribers($subscribersToDelete);
		}

		$report .= sprintf("toAdd:\n%s\ntoDelete:\n%s\n", print_r($subscribersToAdd, true), print_r($subscribersToDelete, true));
		return $report;
	}

	protected function enqueueReceivers($receivers, $operation)
	{
		if(!in_array($operation, array('ADD', 'DEL')))
		{
			throw new Exception('invalid operation');
		}
		foreach($receivers as $email)
		{
			$this->db->insert('cleverreach_api_sync',
			                  array(
			                  	'customers_id' => 0,
			                  	'email' => $email,
			                  	'action' => $operation,
			                  	'status' => 0,
			                  	'created_at' => date('Y-m-d H:i:s'),
			                  )
			                 );
		}
		$this->db->query(
			'UPDATE `cleverreach_api_sync` cas
			LEFT OUTER JOIN customers c ON c.customers_email_address = cas.email
			SET cas.customers_id = c.customers_id
			WHERE cas.status = 0');
	}

	protected function getRemoteSubscribers(CleverReachGroup $group)
	{
		$remoteSubscribers = array();
		$pagesize = 5000;
		$type = 'active';
		$page = 0;
		$remoteSubscribersPage = $group->getReceivers($page, $pagesize, $type);
		while(!empty($remoteSubscribersPage))
		{
			foreach($remoteSubscribersPage as $remoteSubscriber)
			{
				$remoteSubscribers[] = $remoteSubscriber->email;
			}
			$page++;
			$remoteSubscribersPage = $group->getReceivers($page, $pagesize, $type);
		}
		return $remoteSubscribers;
	}

	public function getLocalSubscribers()
	{
		$localSubscribers = array();
		$customersQuery = $this->db->select('customers_email_address')
		                           ->where('customers_newsletter', 1)
		                           ->get('customers');
		foreach($customersQuery->result() as $customerRow)
		{
			$localSubscribers[] = $customerRow->customers_email_address;
		}

		$recipientsQuery = $this->db->select('customers_email_address')
		                            ->where('mail_status', 1)
		                            ->get('newsletter_recipients');
		foreach($recipientsQuery->result() as $customerRow)
		{
			$localSubscribers[] = $customerRow->customers_email_address;
		}
		$localSubscribers = array_unique($localSubscribers);
		return $localSubscribers;
	}

	protected function activateLocalSubscribers($subscribers)
	{
		require_once DIR_FS_INC . 'xtc_random_charcode.inc.php';

		foreach($subscribers as $email)
		{
			$customer = $this->getCustomerByEmail(MainFactory::create('CustomerEmail', $email));
			if($customer === false)
			{
				continue;
			}
			$this->db->set('customers_newsletter', 1)
			     ->where('customers_id', (int)$customer->getId())
			     ->update('customers');

			$newsletterRecipientsCheck =
				$this->db->select('mail_id')
				         ->where('customers_email_address', $email)
				         ->get('newsletter_recipients');
			if($newsletterRecipientsCheck->num_rows() > 0)
			{
				$newsletterRecipient = $newsletterRecipientsCheck->row();
				$this->db->where('mail_id', $newsletterRecipient->mail_id)
				         ->set('mail_status', 1)
				         ->update('newsletter_recipients');
			}
			else
			{
				$newsletterRecipientData = array(
					'customers_email_address' => $email,
					'customers_id'            => $customer->getId(),
					'customers_status'        => $customer->getStatusId(),
					'customers_firstname'     => $customer->getFirstname(),
					'customers_lastname'      => $customer->getLastname(),
					'mail_status'             => '1',
					'mail_key'                => xtc_random_charcode(32),
					'date_added'              => date('Y-m-d H:i:s'),
				);
				$this->db->insert('newsletter_recipients', $newsletterRecipientData);
			}
		}
	}

	protected function deactivateLocalSubscribers($subscribers)
	{
		foreach($subscribers as $email)
		{
			$customer = $this->getCustomerByEmail(MainFactory::create('CustomerEmail', $email));
			if($customer !== false)
			{
				$this->db->where('customers_id', $customer->getId())
				         ->set('customers_newsletter', 0)
				         ->update('customers');
				$this->db->where('customers_id', $customer->getId())
				         ->delete('newsletter_recipients');
			}
			else
			{
				$this->db->where('customers_email_address', $email)
				         ->delete('newsletter_recipients');
			}
		}
	}

	protected function getCustomerByEmail(CustomerEmail $customerEmail)
	{
		$customerReadService = StaticGXCoreLoader::getService('CustomerRead');
		$customers = $customerReadService->filterCustomers(array('customers_email_address' => (string)$customerEmail));
		$customer = false;
		if(count($customers) === 1)
		{
			$customer = $customers[0];
		}
		return $customer;
	}

	/**
	 * returns a DateTime representation of customers.customer_date_added
	 *
	 * This is a stand-in for the hitherto missing CustomerInterface::getDateAdded()
	 *
	 * @param $customer CustomerInterface
	 * @return DateTime date added
	 */
	protected function getCustomerDateAdded(CustomerInterface $customer)
	{
		$db = StaticGXCoreLoader::getDatabaseQueryBuilder();
		$query = $db->select('customers_date_added')->where('customers_id', (int)$customer->getId())->get('customers');
		$row = $query->row();
		$customerDateAdded = new DateTime($row->customers_date_added);
		return $customerDateAdded;
	}

}