<?php
/* --------------------------------------------------------------
	CleverReachGroup.inc.php 2016-04-11
	Gambio GmbH
	http://www.gambio.de
	Copyright (c) 2015 Gambio GmbH
	Released under the GNU General Public License (Version 2)
	[http://www.gnu.org/licenses/gpl-2.0.html]
	--------------------------------------------------------------
*/

class CleverReachGroup
{
	protected $metadata;
	protected $configStorage;
	protected $requiredAttributes = array('firstname', 'lastname', 'street', 'zip', 'city', 'country', 'company');

	public function __construct($groupId = null)
	{
		$this->configStorage = MainFactory::create('CleverReachConfigurationStorage');
		if($groupId === null)
		{
			$groupId = $this->configStorage->get('newsletter_group_id');
		}
		if((int)$groupId <= 0)
		{
			throw new Exception('Group ID must be given or newsletter_group_id must be configured');
		}
		$cleverReachRestService = MainFactory::create('CleverReachRestService');
		$cleverReachRestRequest = MainFactory::create('CleverReachRestRequest', 'GET', '/groups.json/'.$groupId);
		$response = $cleverReachRestService->performRequest($cleverReachRestRequest);
		if($response->getResponseCode() !== 200)
		{
			throw new Exception('Invalid group '.$groupId);
		}
		$this->metadata = $response->getResponseObject();
	}

	public function __get($name)
	{
		$value = null;
		$methodName = 'get'.ucfirst($name);
		if(method_exists($this, $methodName))
		{
			$value = $this->$methodName();
		}
		else
		{
			if(property_exists($this->metadata, $name))
			{
				$value = $this->metadata->$name;
			}
		}
		return $value;
	}

	public function getReceivers($page = 0, $pagesize = 5000, $type = 'all')
	{
		$parameters = array(
			'page' => (int)$page,
			'pagesize' => (int)$pagesize,
			'type' => $type,
		);
		$cleverReachRestService = MainFactory::create('CleverReachRestService');
		$cleverReachRestRequest = MainFactory::create('CleverReachRestRequest', 'GET', '/groups.json/'.$this->id.'/receivers?'.http_build_query($parameters));
		$response = $cleverReachRestService->performRequest($cleverReachRestRequest);
		if($response->getResponseCode() !== 200)
		{
			throw new Exception('Invalid group '.$groupId);
		}
		return $response->getResponseObject();
	}

	public function addReceiver($emailAddress, $additionalData = null)
	{
		$receiverData = array(
			'email' => $emailAddress,
		);
		if(is_array($additionalData))
		{
			$receiverData = array_merge($receiverData, $additionalData);
		}

		$postData = array($receiverData);
		$this->addReceivers($postData);
	}

	/**
	 * adds receivers in bulk
	 * @param array $receivers receivers
	 */
	public function addReceivers($receivers)
	{
		$cleverReachRestService = MainFactory::create('CleverReachRestService');
		$cleverReachRestRequest = MainFactory::create('CleverReachRestRequest', 'POST', '/groups.json/'.$this->id.'/receivers', json_encode($receivers));
		$response = $cleverReachRestService->performRequest($cleverReachRestRequest);
		if($response->getResponseCode() !== 200)
		{
			throw new Exception('Error adding new receivers to group '.$this->id);
		}
	}

	public function deleteReceiver($emailAddress)
	{
		$cleverReachRestService = MainFactory::create('CleverReachRestService');
		$cleverReachRestRequest = MainFactory::create('CleverReachRestRequest', 'DELETE', '/groups.json/'.$this->id.'/receivers/'.urlencode($emailAddress));
		$response = $cleverReachRestService->performRequest($cleverReachRestRequest);
		if($response->getResponseCode() !== 200)
		{
			throw new Exception('Error deleting receiver '.$emailAddress.' from group '.$this->id);
		}
		return $response->getResponseObject();
	}

	public function getAttributes()
	{
		$cleverReachRestService = MainFactory::create('CleverReachRestService');
		$cleverReachRestRequest = MainFactory::create('CleverReachRestRequest', 'GET', '/groups.json/'.$this->id.'/attributes');
		$response = $cleverReachRestService->performRequest($cleverReachRestRequest);
		if($response->getResponseCode() !== 200)
		{
			throw new Exception(sprintf('Failed to retrieve attributes for group %d', $this->id));
		}
		return $response->getResponseObject();
	}

	/**
	 * Adds an attribute to the group
	 * @param string $name attribute name
	 * @param string $type attribute type (text/number/gender/date)
	 * @return stdClass JSON response, decoded
	 */
	public function addAttribute($name, $type = 'text', $description = '', $preview_value = '')
	{
		$name          = (string)$name;
		$type          = (string)$type;
		$description   = (string)$description;
		$preview_value = (string)$preview_value;
		if(!in_array($type, array('text', 'number', 'gender', 'date')))
		{
			throw new Exception('invalid attribute type ' . $type);
		}
		$postData = array(
			'name' => $name,
			'type' => $type,
		);
		if(!empty($description))
		{
			$postData['description'] = $description;
		}
		if(!empty($preview_value))
		{
			$postData['preview_value'] = $preview_value;
		}
		$cleverReachRestService = MainFactory::create('CleverReachRestService');
		$cleverReachRestRequest = MainFactory::create('CleverReachRestRequest', 'POST', '/groups.json/'.$this->id.'/attributes', json_encode($postData));
		$response = $cleverReachRestService->performRequest($cleverReachRestRequest);
		if($response->getResponseCode() !== 200)
		{
			throw new Exception(sprintf('Could not add attribute %s to group %d', $name, $this->id));
		}
		return $response->getResponseObject();
	}

	/**
	 * adds required attributes to group (if not already present)
	 * @return array List of added attributes
	 */
	public function addRequiredAttributes()
	{
		$missingAttributes = $this->requiredAttributes;
		$attributes = $this->getAttributes() ?: array();
		$attributesNames = array();
		foreach($attributes as $attribute)
		{
			$attributesNames[$attribute->id] = $attribute->name;
			$missingAttributes = array_diff($missingAttributes, array($attribute->name));
		}
		$addedAttributes = $missingAttributes;
		foreach($missingAttributes as $missingAttribute)
		{
			$this->addAttribute($missingAttribute, 'text');
		}
		return $addedAttributes;
	}

	public function addOrder($emailAddress, $ordersId, $productName, $productId, $timeStamp, $price, $quantity, $mailingId, $storeName, $currency)
	{
		$postData = array(
			'order_id' => $ordersId,
			'product' => $productName,
			'product_id' => $productId,
			'stamp' => (int)$timeStamp,
			'price' => (double)$price,
			'quantity' => (int)round($quantity),
			'mailing_id' => (int)$mailingId,
			'source' => $storeName,
			'currency' => $currency,
		);
		$cleverReachRestService = MainFactory::create('CleverReachRestService');
		$serviceUrl = sprintf('/groups.json/%d/receivers/%s/orders', $this->id, urlencode($emailAddress));
		$cleverReachRestRequest = MainFactory::create('CleverReachRestRequest', 'POST', $serviceUrl, json_encode($postData));
		$response = $cleverReachRestService->performRequest($cleverReachRestRequest);
		return $response->getResponseObject();
	}

}